from django.conf import settings
from django.db.models import Prefetch
from rest_framework import serializers

from arkindex.documents.models import Element, MetaData, MetaType
from arkindex.images.models import Image
from arkindex.project.tools import bounding_box, build_absolute_url


class ImageResourceManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an image into a IIIF resource.
    """

    def to_representation(self, image):
        assert isinstance(image, Image)
        return {
            "@id": image.get_thumbnail_url(max_width=None, max_height=None),
            "@type": "dctypes:Image",
            "height": image.height,
            "width": image.width,
            "service": {
                "@context": settings.IIIF_IMAGE_CONTEXT,
                "@id": image.url,
                "profile": settings.IIIF_IMAGE_SERVICE_PROFILE
            }
        }


class ImageThumbnailManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an image into a IIIF manifest thumbnail
    """

    def to_representation(self, image):
        assert isinstance(image, Image)
        return {
            "@id": image.get_thumbnail_url(),
            "service": {
                "@context": settings.IIIF_IMAGE_CONTEXT,
                "@id": image.url,
                "profile": settings.IIIF_IMAGE_SERVICE_PROFILE
            }
        }


class ManifestMetadataSerializer(serializers.ModelSerializer):

    label = serializers.CharField(source="name")

    class Meta:
        model = MetaData
        fields = ("label", "value")


class ElementCanvasManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an element's zone into a IIIF canvas
    """

    def get_other_content(self, element):
        return []

    def to_representation(self, element):
        assert isinstance(element, Element)
        assert element.image is not None, "Element has no image"
        assert element.polygon is not None, "Element has no polygon"
        assert "request" in self.context, "A request is required to generate absolute URLs"

        if element.folio_metadatas:
            label = element.folio_metadatas[0].value
        else:
            label = element.name

        _, _, width, height = bounding_box(element.polygon)

        return {
            "@id": build_absolute_url(element, self.context["request"], "api:iiif-canvas"),
            "@type": "sc:Canvas",
            "label": label,
            "height": int(height),
            "width": int(width),
            "images": [
                {
                    "@type": "oa:Annotation",
                    "resource": ImageResourceManifestSerializer(element.image, context=self.context).data,
                    "on": build_absolute_url(element, self.context["request"], "api:iiif-canvas"),
                    "motivation": "sc:painting"
                }
            ],
            "otherContent": self.get_other_content(element),
        }


class PageCanvasManifestSerializer(ElementCanvasManifestSerializer):
    """
    Serialize a page's zone into a IIIF canvas
    """

    def get_other_content(self, page):
        return [
            {
                "@id": build_absolute_url(page, self.context["request"], "api:element-annotation-list"),
                "@type": "sc:AnnotationList",
                "label": "Transcriptions",
            }
        ]


class FolderManifestSerializer(serializers.Serializer):
    """
    Serialize a folder into a IIIF manifest
    """

    canvas_serializer = PageCanvasManifestSerializer
    id_url_name = "api:folder-manifest"

    def to_representation(self, element):
        assert isinstance(element, Element)
        assert "request" in self.context, "A request is required to generate absolute URLs"

        canvas_elements = self.get_canvases(element)
        canvases = self.canvas_serializer(
            canvas_elements,
            context=self.context,
            many=True
        ).data
        if len(canvas_elements) > 0:
            thumbnail = ImageThumbnailManifestSerializer(canvas_elements[0].image).data
        else:
            thumbnail = None

        return {
            "@context": settings.IIIF_PRESENTATION_CONTEXT,
            "@id": build_absolute_url(element, self.context["request"], self.id_url_name),
            "@type": "sc:Manifest",
            "thumbnail": thumbnail,
            "related": [],
            "structures": self.get_structures(element, canvases),
            "description": "",
            "sequences": [
                {
                    "canvases": canvases,
                    "label": "",
                    "@id": build_absolute_url(element, self.context["request"], "api:iiif-sequence"),
                    "@type": "sc:Sequence"
                }
            ],
            "viewingHint": "individuals",
            "label": element.name,
            "viewingDirection": "left-to-right",
            "service": [],
            "metadata": ManifestMetadataSerializer(
                element.metadatas.exclude(type=MetaType.Markdown),
                context=self.context,
                many=True,
            ).data,
        }

    def get_canvases(self, element):
        assert isinstance(element, Element) and element.type.folder

        folio_prefetch = Prefetch(
            "metadatas",
            MetaData.objects.filter(name="folio"),
            to_attr="folio_metadatas",
        )

        return Element.objects \
            .filter(
                paths__path__last=element.id,  # Direct children
                type__folder=False,
                image_id__isnull=False,
                polygon__isnull=False,
            ).order_by("paths__ordering") \
            .prefetch_related("image__server", folio_prefetch)

    def get_structures(self, element, canvases):
        return [{
            "viewingHint": "top",
            "label": element.name,
            "@id": build_absolute_url(element, self.context["request"], self.id_url_name),
            "ranges": [c["@id"] for c in canvases],
            "@type": "sc:Range"
        }] + [
            {
                "canvases": [c["@id"]],
                "label": c["label"],
                "@id": c["@id"],
                "within": c["@id"],
                "@type": "sc:Range"
            }
            for c in canvases]
