from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.images.models import ImageServer
from arkindex.process.models import WorkerRun
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestPatchElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug="volume")
        cls.page_type = cls.corpus.types.get(slug="page")
        cls.line_type = cls.corpus.types.get(slug="text_line")
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.element = Element.objects.get(name="Volume 1, page 2r")
        cls.line = cls.corpus.elements.get(name="Text line")
        cls.line_2 = Element.objects.create(
            corpus=cls.corpus,
            name="Line 2",
            type=cls.line_type,
            image=cls.line.image,
            polygon=((0, 0), (0, 12), (10, 12), (10, 0), (0, 0)),
        )
        cls.image = ImageServer.objects.local.images.create(
            path="kingdom/far/away",
            status=S3FileStatus.Checked,
            width=42,
            height=42,
        )
        cls.private_corpus = Corpus.objects.create(name="private", public=False)
        cls.private_elt = cls.private_corpus.elements.create(type=cls.private_corpus.types.create(slug="type"))

    def test_patch_element_unverified(self):
        """
        Check patching an element with an unverified email is not possible
        """
        user = User.objects.create_user("nope@nope.fr")
        self.assertFalse(user.verified_email)
        self.client.force_login(user)

        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}),
            data={"name": "Untitled (2)"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have permission to perform this action."}
        )

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_patch_no_write_access(self, has_access_mock):
        # Create read_only corpus right
        self.private_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        self.assertTrue(self.user.verified_email)
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.private_elt.id)}),
            data={"name": "Untitled (2)"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {"detail": "You do not have write access to this element."}
        )

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.private_corpus, Role.Contributor.value, skip_public=False))

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_patch_element_no_read_access(self, filter_rights_mock):
        """
        Check patching an element as anonymous user is not possible
        """
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.private_elt.id)}),
            data={"name": "Untitled (2)"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    def test_patch_element(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}),
                data={"name": "Untitled (2)"},
                format="json",
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["name"], "Untitled (2)")
        self.vol.refresh_from_db()
        self.assertEqual(self.vol.name, "Untitled (2)")

    def test_patch_element_name_type_only(self):
        """
        Ensure patching an element only updates name and type, nothing else,
        and every other property is ignored
        """
        self.client.force_login(self.user)
        box_type = self.corpus.types.create(slug="xyzzy", display_name="box", folder=True)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}),
            data={
                "id": "beef",
                "name": "New name",
                "type": "xyzzy",
                "corpus": "something",
                "thumbnail_url": "/dev/urandom",
                "thumbnail_put_url": "/dev/null",
                "zone": 51,
                "metadata": [],
                "classifications": [],
                "worker_run": str(WorkerRun.objects.first().id),
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["name"], "New name")
        self.assertEqual(data["type"], "xyzzy")
        self.assertIsNone(data["worker_run"])
        self.vol.refresh_from_db()
        self.assertEqual(self.vol.name, "New name")
        self.assertEqual(self.vol.type, box_type)
        self.assertNotEqual(self.vol.id, "beef")
        self.assertNotEqual(self.vol.corpus_id, "something")
        self.assertIsNone(self.vol.worker_version_id)
        self.assertIsNone(self.vol.worker_run_id)

    def test_patch_element_polygon(self):
        self.client.force_login(self.user)
        self.assertIsNotNone(self.element.image_id)
        self.assertIsNotNone(self.element.polygon)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.element.id)}),
            data={
                "polygon": [[10, 20], [10, 30], [30, 30], [30, 20], [10, 20]]
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["name"], self.element.name)
        self.assertEqual(response.json()["type"], self.element.type.slug)
        self.element.refresh_from_db()
        self.assertTupleEqual(
            self.element.polygon.coords,
            ((10, 20), (10, 30), (30, 30), (30, 20), (10, 20))
        )

    def test_patch_element_polygon_without_image(self):
        self.client.force_login(self.user)
        self.vol.image = None
        self.vol.polygon = None
        self.vol.save()
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}),
            data={
                "polygon": [[10, 20], [10, 30], [30, 30], [30, 20], [10, 20]]
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "image": ["Image is required when defining a polygon on an element without an existing zone"]
        })

    def test_patch_element_image_preserve_polygon(self):
        self.client.force_login(self.user)
        self.assertNotEqual(self.line_2.image, self.image)
        expected_polygon = self.line_2.polygon.clone()
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.line_2.id)}),
            data={
                "image": str(self.image.id),
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.line_2.refresh_from_db()
        self.assertEqual(self.line_2.image, self.image)
        self.assertEqual(self.line_2.polygon, expected_polygon)

    def test_patch_element_new_zone(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.vol.image_id)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}),
            data={
                "image": str(self.image.id),
                "polygon": [[10, 20], [10, 30], [30, 30], [30, 20], [10, 20]],
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.vol.refresh_from_db()
        self.assertEqual(self.vol.image, self.image)
        self.assertTupleEqual(
            self.vol.polygon.coords,
            ((10, 20), (10, 30), (30, 30), (30, 20), (10, 20))
        )

    def test_patch_element_invalid_dimensions(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.vol.image)
        bad_image = self.imgsrv.images.create(
            path="oh-no",
            status=S3FileStatus.Unchecked,
            width=0,
            height=0,
        )
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.vol.id)}),
            data={
                "image": str(bad_image.id),
                "polygon": [[10, 20], [10, 30], [30, 30], [30, 20], [10, 20]],
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"image": ["This image does not have valid dimensions."]})
        self.assertIsNone(self.vol.image)

    def test_patch_element_polygon_negative(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.element.id)}),
            data={
                "polygon": [[-10, -20], [10, 30], [30, 30], [30, 20], [10, 20]]
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "polygon": {
                "0": {
                    "0": ["Ensure this value is greater than or equal to 0."],
                    "1": ["Ensure this value is greater than or equal to 0."],
                }
            }
        })

    def test_patch_element_polygon_outside_image(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.element.id)}),
            data={
                "polygon": [[0, 1009], [10, 30], [30, 30], [30, 20], [10, 20]]
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "polygon": ["An element's polygon must not exceed its image's bounds."]
        })

    def test_patch_element_null_polygon_null_image(self):
        self.assertIsNotNone(self.element.image)
        self.assertIsNotNone(self.element.polygon)
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.element.id)}),
            data={
                "polygon": None,
                "image": None,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.element.refresh_from_db()
        self.assertIsNone(self.element.image)
        self.assertIsNone(self.element.polygon)

    def test_patch_element_image_only(self):
        self.element.image = None
        self.element.polygon = None
        self.element.save()
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:element-retrieve", kwargs={"pk": str(self.element.id)}),
            data={"image": str(self.image.id)},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.element.refresh_from_db()
        self.assertEqual(self.element.image_id, self.image.id)
        self.assertTupleEqual(
            self.element.polygon.coords,
            ((0, 0), (0, 42), (42, 42), (42, 0), (0, 0)),
        )
