import uuid
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element, ElementPath
from arkindex.images.models import Image
from arkindex.process.models import Process, ProcessMode
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestProcessElements(FixtureAPITestCase):
    """
    Test list process elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        # ----------------- TREE ------------------
        #
        #        Folder 1          Folder 2
        #          /|\               /|\
        #         / | \             / | \
        #        /  |  \           /  |  \
        #       /   |   \         /   |   \
        #      /    |    \       /    |    \
        #     /     |     \     /     |     \
        #    /      |      \   /      |      \
        #   /       |       \ /       |       \
        # Page 1  Page 2   Page 3  Page 4  Page 5
        #   |      |         |       |       |
        # Line 1  Line 2   Line 3  Line 4  Line 5

        # Create a corpus
        cls.private_corpus = Corpus.objects.create(name="private")

        # Create element types
        cls.folder_type, _ = cls.private_corpus.types.get_or_create(
            slug="folder",
            display_name="Folder",
            folder=True,
        )
        cls.page_type, _ = cls.private_corpus.types.get_or_create(
            slug="page",
            display_name="Page",
        )
        cls.line_type, _ = cls.private_corpus.types.get_or_create(
            slug="text_line",
            display_name="Line",
        )

        # Create 2 folders
        cls.folder_1 = Element.objects.create(
            corpus=cls.private_corpus,
            type=cls.folder_type,
            name="Baba au rhum",
        )
        cls.folder_2 = Element.objects.create(
            corpus=cls.private_corpus,
            type=cls.folder_type,
            name="Montgolfière",
        )
        # Create top-level paths for the folders
        cls.folder_1.paths.create(path=[])
        cls.folder_2.paths.create(path=[])

        # Create 2 pages in each volume and 1 common page
        cls.page_1 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Baba au rhum 1",
            type=cls.page_type
        )
        cls.page_2 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Montgolfière 1",
            type=cls.page_type
        )
        cls.page_3 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Common",
            type=cls.page_type
        )
        cls.page_4 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Montgolfière 2",
            type=cls.page_type,
            rotation_angle=180,
        )
        cls.page_5 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Baba au rhum 2",
            type=cls.page_type,
            mirrored=True,
        )
        cls.page_1.add_parent(cls.folder_1)
        cls.page_2.add_parent(cls.folder_1)
        cls.page_3.add_parent(cls.folder_1)
        cls.page_4.add_parent(cls.folder_2)
        cls.page_5.add_parent(cls.folder_2)
        cls.page_3.add_parent(cls.folder_2)

        # Create 1 line in each page
        cls.line_1 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 1",
            type=cls.line_type,
            image=Image.objects.get(path="img6"),
            polygon=((0, 0), (0, 1000), (1000, 1000), (1000, 0), (0, 0)),
        )
        cls.line_2 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 2",
            type=cls.line_type
        )
        cls.line_3 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 3",
            type=cls.line_type
        )
        cls.line_4 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 4",
            type=cls.line_type
        )
        cls.line_5 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 5",
            type=cls.line_type
        )
        cls.line_1.add_parent(cls.page_1)
        cls.line_2.add_parent(cls.page_2)
        cls.line_3.add_parent(cls.page_3)
        cls.line_4.add_parent(cls.page_4)
        cls.line_5.add_parent(cls.page_5)

        # Add an MLClass to some elements
        cls.ml_class = cls.corpus.ml_classes.create(name="clafoutis")
        cls.line_1.classifications.create(ml_class=cls.ml_class, confidence=.1)
        cls.line_5.classifications.create(ml_class=cls.ml_class, confidence=.2)
        cls.page_4.classifications.create(ml_class=cls.ml_class, confidence=.3)

    def setUp(self):
        super().setUp()
        self.process = Process.objects.create(
            creator_id=self.user.id,
            mode=ProcessMode.Workers,
            corpus_id=self.private_corpus.id
        )

    def test_requires_login(self):
        response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_wrong_id(self):
        self.client.force_login(self.superuser)
        response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": str(uuid.uuid4())}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_no_access(self, has_access_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have an admin access to the corpus of this process."})

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.private_corpus, Role.Admin.value, skip_public=False))

    def test_filter_elements_wrong_corpus(self):
        """
        Selected elements must be part of the same corpus as the process
        """
        self.process.corpus = Corpus.objects.create(name="private")
        self.process.save()
        self.process.elements.add(self.page_1.id, self.folder_2.id)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertDictEqual(response.json(), {
            "__all__": [f"Some elements on this process are not part of corpus {self.process.corpus_id}"]
        })

    def test_filter_elements_multiple_corpus(self):
        """
        Selected elements are part of multiple corpora
        This should not happen in usual situations
        """
        corpus = Corpus.objects.create(name="private")
        element = Element.objects.create(
            corpus=corpus,
            type=self.folder_type,
            name="Folder",
        )
        self.process.elements.add(element.id, self.page_1.id)
        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertDictEqual(response.json(), {
            "__all__": [f"Some elements on this process are not part of corpus {self.process.corpus_id}"]
        })

    def test_filter_element_wrong_corpus(self):
        """
        Element and corpus are different
        This should not happen in usual situations
        """
        self.process.corpus = Corpus.objects.create(name="private")
        self.process.element = self.page_1
        self.process.save()

        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertDictEqual(response.json(), {
            "__all__": [f"Element {self.page_1.id} is not part of corpus {self.process.corpus_id}"]
        })

    def test_filter_name(self):
        self.process.name_contains = "rhum"
        self.process.load_children = True
        self.process.save()
        elements = [self.folder_1, self.page_1, self.page_5]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_filter_name_case_insensitive(self):
        self.process.name_contains = "Rhum"
        self.process.load_children = True
        self.process.save()
        elements = [self.folder_1, self.page_1, self.page_5]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_filter_name_from_element(self):
        self.process.element = self.folder_1
        self.process.name_contains = "rhum"
        self.process.load_children = True
        self.process.save()
        self.process.element = self.folder_1
        self.process.name_contains = "rhum"
        self.process.load_children = True
        self.process.save()

        elements = [self.folder_1, self.page_1]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_filter_type(self):
        self.process.element_type = self.folder_type
        self.process.save()
        elements = [self.folder_1, self.folder_2]

        self.client.force_login(self.superuser)
        with self.assertExactQueries("process_elements_filter_type.sql", skip=1, params={
            "user_id": self.superuser.id,
            "process_id": str(self.process.id),
            "corpus_id": str(self.private_corpus.id),
            "type_id": str(self.folder_type.id),
        }):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_filter_ml_class(self):
        self.process.ml_class = self.ml_class
        self.process.load_children = True
        self.process.save()
        elements = [self.line_1, self.line_5, self.page_4]

        self.client.force_login(self.superuser)
        with self.assertExactQueries("process_elements_filter_ml_class.sql", skip=1, params={
            "user_id": self.superuser.id,
            "process_id": str(self.process.id),
            "corpus_id": str(self.private_corpus.id),
            "ml_class_id": str(self.ml_class.id),
        }):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_filter_type_from_element(self):
        self.process.element = self.folder_1
        self.process.element_type = self.line_type
        self.process.load_children = True
        self.process.save()
        elements = [self.line_1, self.line_2, self.line_3]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_filter_element(self):
        self.process.element = self.page_1
        self.process.save()

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(self.page_1.id),
                "type_id": str(self.page_1.type_id),
                "name": self.page_1.name,
                "confidence": self.page_1.confidence,
            }
        ])

    def test_filter_elements(self):
        self.process.elements.add(self.page_1.id, self.folder_2.id)

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(self.page_1.id),
                "type_id": str(self.page_1.type_id),
                "name": self.page_1.name,
                "confidence": self.page_1.confidence,
            },
            {
                "id": str(self.folder_2.id),
                "type_id": str(self.folder_2.type_id),
                "name": self.folder_2.name,
                "confidence": self.folder_2.confidence,
            }
        ])

    def test_load_children_and_filter_name(self):
        self.process.name_contains = "rhum"
        self.process.load_children = True
        self.process.save()
        elements = [self.folder_1, self.page_1, self.page_5]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_corpus_and_filter_type(self):
        self.process.element_type = self.folder_type
        self.process.load_children = True
        self.process.save()
        elements = [self.folder_1, self.folder_2]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_load_children_and_filter_element(self):
        self.process.element = self.folder_1
        self.process.load_children = True
        self.process.save()
        elements = [self.folder_1, self.page_1, self.page_3, self.line_1, self.line_2, self.line_3, self.page_2]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_load_children_and_filter_elements(self):
        self.process.load_children = True
        self.process.save()
        self.process.elements.add(self.page_1.id, self.folder_2.id)
        elements = [self.page_1, self.page_5, self.page_3, self.line_1, self.line_3, self.line_4, self.line_5, self.folder_2, self.page_4]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(8):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_load_children_and_any_filter(self):
        self.process.load_children = True
        self.process.save()
        elements = Element.objects.filter(corpus=self.private_corpus).order_by("name", "type__slug")

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_non_workers(self):
        self.process.load_children = True
        self.client.force_login(self.superuser)

        # Every mode other than Workers never returns elements
        modes = list(ProcessMode)
        modes.remove(ProcessMode.Workers)
        # Local processes do not have a corpus, causing HTTP 404
        modes.remove(ProcessMode.Local)

        for mode in modes:
            with self.subTest(mode=mode):
                self.process.mode = mode
                self.process.save()
                with self.assertNumQueries(4):
                    response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                data = response.json()
                self.assertDictEqual(data, {
                    "count": None,
                    "next": None,
                    "previous": None,
                    "results": [],
                })

    def test_no_corpus(self):
        """
        ListProcessElements does not support processes without a corpus
        """
        process = self.superuser.processes.get(corpus=None, mode=ProcessMode.Local)
        self.client.force_login(self.superuser)

        process.mode = ProcessMode.Local
        process.save()
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": process.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {"detail": "No Process matches the given query."})

    def test_workers_mode(self):
        # The Workers mode returns some elements
        elements = Element.objects.filter(corpus=self.private_corpus).order_by("id")
        # This tests only supports up to 20 elements
        self.assertLessEqual(elements.count(), 20)
        self.process.corpus = self.private_corpus
        self.process.load_children = True
        self.process.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertIsNone(data["count"])
        self.assertIsNone(data["previous"])
        self.assertIsNone(data["next"])
        self.assertListEqual(data["results"], [
            {
                "id": str(element.id),
                "type_id": str(element.type_id),
                "name": element.name,
                "confidence": element.confidence,
            }
            for element in elements
        ])

    def test_list_elements_cursor_pagination(self):
        """
        No element is duplicated or dropped
        """
        elts = Element.objects.bulk_create([
            Element(
                corpus=self.private_corpus,
                name="Similar name",
                type=self.page_type
            ) for i in range(40)
        ])
        for elt in elts:
            elt.add_parent(self.folder_1)
        self.process.element = self.folder_1
        self.process.name_contains = "Similar"
        self.process.load_children = True
        self.process.save()
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            page_1 = self.client.get(reverse("api:process-elements-list", kwargs={"pk": self.process.id}))
        self.assertEqual(len(page_1.json()["results"]), 20)
        next_page = page_1.json().get("next")
        self.assertIsNotNone(next_page)
        with self.assertNumQueries(7):
            page_2 = self.client.get(next_page)
        self.assertIsNone(page_2.json()["next"])
        qs_1 = Element.objects.filter(id__in=[elt["id"] for elt in page_1.json()["results"]])
        qs_2 = Element.objects.filter(id__in=[elt["id"] for elt in page_2.json()["results"]])

        self.assertEqual(qs_1.intersection(qs_2).count(), 0)
        self.assertEqual(qs_1.union(qs_2).count(), 40)

    def test_elements_invalid_cursor(self):
        self.client.force_login(self.superuser)
        response = self.client.get(
            reverse("api:process-elements-list", kwargs={"pk": self.process.id}),
            {"cursor": "ABC", "with_count": True}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {"detail": "Invalid cursor"})

    def test_elements_count(self):
        """
        Elements count can be retrieved with with_count parameter
        """
        self.client.force_login(self.superuser)
        self.process.load_children = True
        self.process.save()
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:process-elements-list", kwargs={"pk": self.process.id}),
                {"page_size": 6, "with_count": True}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNotNone(data["next"])
        self.assertEqual(data["count"], 12)
        self.assertEqual(len(data["results"]), 6)
        next_url = data.get("next")
        self.assertIn("cursor", next_url)

        second_page = self.client.get(next_url)
        data = second_page.json()
        # Count should be present in the next URL
        self.assertEqual(data["count"], 12)
        self.assertIsNone(data["next"])
        self.assertEqual(len(data["results"]), 6)

    def test_cursor_pagination_page_size(self):
        """
        Page size may be changed for cursor pagination
        """
        elements = [
            Element(
                corpus=self.private_corpus,
                name="Similar name",
                type=self.page_type
            ) for i in range(51)
        ]
        Element.objects.bulk_create(elements)
        ElementPath.objects.bulk_create([
            ElementPath(element=element, path=[], ordering=0)
            for element in elements
        ])
        self.process.name_contains = "Similar"
        self.process.save()

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:process-elements-list", kwargs={"pk": self.process.id}),
                {"page_size": 50}
            )
        data = response.json()
        self.assertEqual(len(data["results"]), 50)
        self.assertIsNotNone(data["next"])

    def test_elements_zone(self):
        """
        The with_image parameter includes zone and image information
        """
        self.client.force_login(self.superuser)
        self.process.load_children = True
        self.process.save()
        with self.assertExactQueries("process_elements_with_image.sql", skip=1, params={
            "user_id": self.superuser.id,
            "process_id": str(self.process.id),
            "corpus_id": str(self.private_corpus.id),
        }):
            response = self.client.get(
                reverse("api:process-elements-list", kwargs={"pk": self.process.id}),
                {"with_image": True}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(self.private_corpus.elements.filter(polygon__isnull=False).exists())
        self.assertCountEqual(response.json()["results"], [
            {
                "id": str(element.id),
                "name": element.name,
                "type_id": str(element.type_id),
                "confidence": element.confidence,
                "image_id": str(element.image_id) if element.image_id else None,
                "image_width": element.image.width if element.image else None,
                "image_height": element.image.height if element.image else None,
                "image_url": element.image.url if element.image else None,
                "polygon": list(map(list, element.polygon)) if element.polygon else None,
                "rotation_angle": element.rotation_angle,
                "mirrored": element.mirrored,
            }
            for element in self.private_corpus.elements.all()
        ])

    def test_corpus_top_level(self):
        self.client.force_login(self.superuser)
        with self.assertExactQueries("process_elements_top_level.sql", skip=1, params={
            "user_id": self.superuser.id,
            "process_id": str(self.process.id),
            "corpus_id": str(self.private_corpus.id),
        }):
            response = self.client.get(
                reverse("api:process-elements-list", kwargs={"pk": self.process.id}),
            )
        self.assertCountEqual(response.json()["results"], [
            {
                "id": str(self.folder_1.id),
                "name": "Baba au rhum",
                "type_id": str(self.folder_1.type_id),
                "confidence": self.folder_1.confidence,
            },
            {
                "id": str(self.folder_2.id),
                "name": "Montgolfière",
                "type_id": str(self.folder_2.type_id),
                "confidence": self.folder_2.confidence,
            }
        ])
